from enum import Enum

import quickfix as fix

from fix_application import FIXApplication, FIXMessage, string_to_message, get_utc_transactime, log


class ExecutionReportType(Enum):
    NewAck = fix.OrdStatus_NEW
    Filled = fix.OrdStatus_FILLED
    DFD = fix.OrdStatus_DONE_FOR_DAY


class ClientApplication(fix.Application):
    UUID = 1234
    session_id = None
    reserve_request_sent = False
    reserve_request_accepted = False
    dfd_sent = False

    def onCreate(self, session_id):
        pass

    def onLogon(self, session_id):
        self.session_id = session_id
        log('CLIENT Session', f"{session_id} logged on.<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<", '\n')
        self.reserve_request_sent = False
        self.reserve_request_accepted = False
        self.dfd_sent = False
        self.send_ioi_query(session_id)

    def onLogout(self, session_id):
        log('CLIENT Session', f"{session_id} logged out.>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>")

    def toAdmin(self, message, session_id):
        message = FIXMessage(message)
        msg_type = message.get_field_value(fix.MsgType())
        if msg_type != fix.MsgType_Heartbeat:
            log('Sent ADMIN', message)

    def fromAdmin(self, message, session_id):
        message = FIXMessage(message)
        msg_type = message.get_field_value(fix.MsgType())
        if msg_type != fix.MsgType_Heartbeat:
            log('Rcvd ADMIN', message)

    def toApp(self, message, session_id):
        log('Sent APP', message)

    def fromApp(self, message, session_id):
        message = FIXMessage(message)
        log('Rcvd APP', message)
        self.process_message(message)

    def process_message(self, message: FIXMessage) -> None:
        msg_type = message.get_field_value(fix.MsgType())
        order_id = message.get_field_value(fix.OrderID())
        if msg_type == fix.MsgType_NewOrderSingle:
            #                msg_type == fix.MsgType_OrderCancelReplaceRequest):
            if message.get_field_value(fix.OrdStatus()) == fix.OrdStatus_NEW:
                self.reserve_request_accepted = True
                log("Reserved ACCEPTED")
            else:
                FIXApplication.set_latest_fix_message_per_order_id(order_id, message)

        elif message.get_field_value(fix.OrdStatus()) == fix.OrdStatus_REJECTED:
            log("Reserved REJECTED")

        elif msg_type == fix.MsgType_OrderCancelRequest:
            FIXApplication.set_latest_fix_message_per_order_id(order_id, None)

    def send_ioi_query(self, session_id):
        message = string_to_message(fix.MsgType_IOI, '|'.join([
            f"28={fix.IOITransType_NEW}",
            f"50={ClientApplication.UUID}"
        ]))
        fix.Session.sendToTarget(message, session_id)

    def send_reserve_request(self, order_id: str, reserve_shares: int):
        if self.reserve_request_sent:
            return

        latest_message = FIXApplication.get_latest_fix_message_per_order_id(order_id)
        if latest_message:
            order_id = latest_message.get_field_value(fix.OrderID())
            clordid = "ITGClOrdID:" + latest_message.get_field_value(fix.OrderID())
            message = string_to_message(fix.MsgType_NewOrderSingle, '|'.join([
                #                f"11={FIXApplication.get_next_clordid()}",
                f"11={clordid}",
                f"37={order_id}",
                f"38={reserve_shares}",
                f"40={latest_message.get_field_value(fix.OrdType())}",
                f"44={latest_message.get_field_value(fix.Price())}",
                f"50={ClientApplication.UUID}",
                f"54={latest_message.get_field_value(fix.Side())}",
                f"55={latest_message.get_field_value(fix.Symbol())}",
                f"60={get_utc_transactime()}",
                f"76={FIXApplication.EXEC_BROKER}",
                f"100={FIXApplication.EX_DESTINATION}",
                f"109={order_id}",
                f"150={fix.ExecType_NEW}",
            ]))

            log("Snd REQUEST", message)
            fix.Session.sendToTarget(message, self.session_id)

            self.reserve_request_sent = True

    def send_execution_report(self, order_id: str, fill_shares: int | None = None,
                              execution_type: ExecutionReportType = ExecutionReportType.NewAck):
        if not self.reserve_request_accepted or self.dfd_sent:
            return

        latest_message = FIXApplication.get_latest_fix_message_per_order_id(order_id)
        if latest_message:
            order_qty = int(latest_message.get_field_value(fix.OrderQty()))
            order_id = latest_message.get_field_value(fix.OrderID())
            clordid = latest_message.get_field_value(fix.ClOrdID())
            expire_time = get_utc_transactime(5 * 60)  # expire 5 mins from now
            price = '11.22'
            cum_qty = fill_shares
            if execution_type == ExecutionReportType.NewAck:
                price = '0'
                cum_qty = '0'
                last_px = '0'
                last_shares = '0'
                order_status = fix.OrdStatus_NEW
                exec_type = fix.ExecType_NEW
                log_msg_type = 'Snd ACK'
                leaves_qty = order_qty
            elif execution_type == ExecutionReportType.DFD:
                last_px = '0'
                last_shares = '0'
                order_status = fix.OrdStatus_DONE_FOR_DAY
                exec_type = fix.ExecType_DONE_FOR_DAY
                log_msg_type = 'Snd DFD'
                leaves_qty = 0
            else:
                if fill_shares is None:
                    fill_shares = order_qty
                elif fill_shares == 0:
                    # nothing to fill. skip this report
                    return
                assert order_qty >= fill_shares, f"fill_shares:{fill_shares} > order's qty:{order_qty}"

                last_px = price
                last_shares = fill_shares
                if order_qty == fill_shares:
                    exec_type = fix.ExecType_FILL
                    order_status = fix.OrdStatus_FILLED
                    log_msg_type = 'Snd Fill'
                else:
                    exec_type = fix.ExecType_PARTIAL_FILL
                    order_status = fix.OrdStatus_PARTIALLY_FILLED
                    log_msg_type = 'Snd Partial'
                leaves_qty = int(order_qty) - fill_shares

            message = string_to_message(fix.MsgType_ExecutionReport, '|'.join([
                f"6={price}",
                f"11={clordid}",
                f"14={cum_qty}",
                f"15={FIXApplication.CURRENCY}",
                f"17={order_id}-fill",
                f"20={fix.ExecTransType_NEW}",
                f"29={fix.LastCapacity_AGENT}",
                f"30={FIXApplication.LAST_MARKET}",
                f"31={last_px}",
                f"32={last_shares}",
                f"37={order_id}",
                f"38={order_qty}",
                f"39={order_status}",
                f"40={latest_message.get_field_value(fix.OrdType())}",
                f"41={clordid}",
                f"47={fix.Rule80A_AGENCY_SINGLE_ORDER}",
                f"50={ClientApplication.UUID}",
                f"54={latest_message.get_field_value(fix.Side())}",
                f"55={latest_message.get_field_value(fix.Symbol())}",
                f"59={latest_message.get_field_value(fix.TimeInForce())}",
                f"60={get_utc_transactime()}",
                f"76={FIXApplication.EXEC_BROKER}",
                f"126={expire_time}",
                f"150={exec_type}",
                f"151={leaves_qty}",
            ]))

            log(log_msg_type, message)
            fix.Session.sendToTarget(message, self.session_id)

            if execution_type == ExecutionReportType.DFD:
                self.dfd_sent = True



import json
import os
from datetime import datetime
from typing import Dict, List

import pandas as pd
from pandas import DataFrame, Series


class OrderManager:
    ORDERS_FILE_PATH = 'oms_orders.csv'
    ORDER_CHANGES_FILE_PATH = 'oms_order_changes.json'
    ORDER_CHANGES_TMP_FILE_PATH = 'oms_order_changes.json.tmp'
    COLUMN_DTYPE_PER_NAME = {
        'order_id': 'string',
        'is_active': 'bool',
        'uuid': 'Int32',
        'symbol': 'string',
        'side': 'category',
        'shares': 'Int32',
        'price': 'Float32',
    }
    SIDES = {'Buy': 1, 'Sell': 2, 'Short': 5}

    def __init__(self):
        self.orders_df = None
        self.last_order_changes_timestamp = datetime.now()
        self.read_orders_from_file()

    def read_orders_from_file(self) -> DataFrame:
        self.orders_df = pd.read_csv(OrderManager.ORDERS_FILE_PATH)

        for name, dtype in OrderManager.COLUMN_DTYPE_PER_NAME.items():
            self.orders_df[name] = self.orders_df[name].astype(dtype)

        return self.orders_df

    def save_orders(self):
        self.orders_df.to_csv(OrderManager.ORDERS_FILE_PATH, index=False)

    def get_orders_for_uuid(self, uuid: str) -> List[Series]:
        orders: List[Series] = []
        for index, row in self.orders_df.iterrows():
            if row['is_active'] and str(row['uuid']) == uuid:
                orders.append(row)

        return orders

    def save_order_change_instructions(self, order_changes: Dict[str, Dict[str, str]]):
        with open(OrderManager.ORDER_CHANGES_TMP_FILE_PATH, "w") as fp:
            json.dump(order_changes, fp, indent=4)

        os.rename(OrderManager.ORDER_CHANGES_TMP_FILE_PATH, OrderManager.ORDER_CHANGES_FILE_PATH)

    def check_and_process_order_change_instructions(self):
        file_mod_time = os.path.getmtime(OrderManager.ORDER_CHANGES_FILE_PATH)
        file_mod_datetime = datetime.fromtimestamp(file_mod_time)

        if file_mod_datetime > self.last_order_changes_timestamp:
            print("Detected order changes...")
            self.last_order_changes_timestamp = file_mod_datetime
            with open(OrderManager.ORDER_CHANGES_FILE_PATH, "r") as fp:
                order_changes = json.load(fp)
                self.process_order_changes(order_changes)

    def process_order_changes(self, order_changes: Dict[str, Dict[str, str]]):
        # the passed changes are tightly bound with streamlit's st.session_state after changing data in the data_editor
        # Example:
        # {
        #   "edited_rows": {
        #     "1": {
        #       "is_active": false
        #     },
        #     "2": {
        #       "symbol": "AAA"
        #     }
        #   },
        #   "added_rows": [
        #     {
        #       "order_id": "111",
        #       "is_active": true,
        #       "uuid": 888,
        #       "symbol": "DDDD",
        #       "side": "Buy",
        #       "shares": 1999,
        #       "price": 99.99
        #     }
        #   ],
        #   "deleted_rows": [
        #     5
        #   ]
        # }

        # It assumes that the changes have already been made and are already on the order file
        order_df = self.read_orders_from_file()

        edited_rows = order_changes["edited_rows"]
        for index, changes in edited_rows.items():
            order_row = order_df.iloc[int(index)]
            self.process_edited_added_row(order_row, changes, True)

        added_rows = order_changes["added_rows"]
        for added_row in added_rows:
            order_id = added_row['order_id']
            order_rows_for_order_id = order_df[order_df['order_id'] == order_id]
            if len(order_rows_for_order_id) >= 1:
                order_row = order_rows_for_order_id.iloc[0]
                self.process_edited_added_row(order_row, added_row, False)
            else:
                print(f"Can't find added row with order_id:{order_id}")

    def process_edited_added_row(self, order_row: Series, changes: Dict[str, str], is_edited: bool):
        # import here to avoid circular import dependencies
        from server_application import ServerApplication, MessageAction

        uuid = order_row['uuid']
        if ServerApplication.is_uuid_of_interest(uuid):
            if is_edited:
                if 'is_active' in changes:
                    message_action = MessageAction.NewOrder if changes['is_active'] else MessageAction.CancelOrder
                else:
                    message_action = MessageAction.ChangeOrder
            else:
                message_action = MessageAction.NewOrder

            ServerApplication.create_order_message(message_action, order_row, True)
        else:
            print(f"Changes requested for uuid:{uuid} but no interest there")



import queue
from enum import Enum
from typing import Set

import quickfix as fix
from pandas import Series

from fix_application import FIXApplication, FIXMessage, get_utc_transactime, log, string_to_message, \
    create_fix_string_from_dict
from order_manager import OrderManager


class MessageAction(Enum):
    NewOrder = fix.MsgType_NewOrderSingle
    ChangeOrder = fix.MsgType_OrderCancelReplaceRequest
    CancelOrder = fix.MsgType_OrderCancelRequest
    RejectOrder = fix.MsgType_ExecutionReport


class ServerApplication(fix.Application):
    order_manager = None
    session_id = None
    uuids_of_interest: Set[str] = set()

    def __init__(self):
        super().__init__()
        self.order_manager = OrderManager()
        self.from_app_queue = queue.Queue()

    def onCreate(self, session_id):
        pass

    def onLogon(self, session_id):
        ServerApplication.session_id = session_id
        log('SERVER Session',
            f"{session_id} logged on.<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<", '\n')

    def onLogout(self, session_id):
        log('SERVER Session',
            f"{session_id} logged out.>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>", '\n')

    def toAdmin(self, message, session_id):
        message = FIXMessage(message)
        msg_type = message.get_field_value(fix.MsgType())
        if msg_type != fix.MsgType_Heartbeat:
            log('Sent ADMIN', message)

    def fromAdmin(self, message, session_id):
        message = FIXMessage(message)
        msg_type = message.get_field_value(fix.MsgType())
        if msg_type != fix.MsgType_Heartbeat:
            log('Rcvd ADMIN', message)

    def toApp(self, message, session_id):
        log('Sent APP', message)

    def fromApp(self, message, session_id):
        message = FIXMessage(message)
        log('Rcvd APP', message)
        self.process_message(message)

    def process_message_from_app_queue(self):
        if self.from_app_queue.not_empty:
            try:
                self.from_app_queue.get(block=True, timeout=1)
            except queue.Empty:
                return

    def check_for_order_changes(self):
        self.order_manager.check_and_process_order_change_instructions()

    def process_message(self, message: FIXMessage) -> None:
        msg_type = message.get_field_value(fix.MsgType())
        if msg_type == fix.MsgType_IOI:
            self.process_ioi_message(message)
        elif msg_type == fix.MsgType_NewOrderSingle:
            self.process_reserve_request_message(message)
        elif msg_type == fix.MsgType_ExecutionReport:
            self.process_execution_report_message(message)

    def process_ioi_message(self, message: FIXMessage):
        uuid = message.get_field_value(fix.SenderSubID())
        ServerApplication.uuids_of_interest.add(uuid)
        uuid_orders = self.order_manager.get_orders_for_uuid(uuid)
        for order in uuid_orders:
            ServerApplication.create_order_message(MessageAction.NewOrder, order, True, True)

    def process_reserve_request_message(self, message: FIXMessage):
        order_id = message.get_field_value(fix.OrderID())
        # odd/even order_id heuristic to decide whether to accept or reject request
        if int(order_id[-1]) % 2 == 0:
            # Before sending the accept first send a 35=G with the reduced qty
            latest_message = FIXApplication.get_latest_fix_message_per_order_id(order_id)
            if latest_message:
                qty_to_reserve = int(message.get_field_value(fix.OrderQty()))
                corrected_qty = int(latest_message.get_field_value(fix.OrderQty())) - qty_to_reserve
                self.send_correct_message(order_id, corrected_qty)

                self.send_reserve_accept_message(message)
        else:
            latest_message = FIXApplication.get_latest_fix_message_per_order_id(order_id)
            if latest_message:
                self.send_reserve_reject_message(message)

    def process_execution_report_message(self, message: FIXMessage):
        # For now, only do something once we get the DFD
        if message.get_field_value(fix.OrdStatus()) == fix.OrdStatus_DONE_FOR_DAY:
            order_id = message.get_field_value(fix.OrderID())
            # figure out the new qty
            cum_qty = int(message.get_field_value(fix.CumQty()))
            order_qty = int(message.get_field_value(fix.OrderQty()))
            new_qty = order_qty - cum_qty
            self.send_correct_message(order_id, int(new_qty))
            # TODO: update csv file

    def send_correct_message(self, order_id: str, corrected_qty: int = 0):
        correct_message = FIXApplication.get_latest_fix_message_per_order_id(order_id)
        if correct_message:
            correct_message.set_field_value(fix.OrderQty(), str(corrected_qty))
            correct_message.set_field_value(fix.OrdStatus(), None)
            ServerApplication.create_order_message(MessageAction.ChangeOrder, correct_message, True, True)

    def send_reserve_accept_message(self, reserve_request_message: FIXMessage):
        reserve_accept_message = FIXMessage(reserve_request_message)
        # Only (un)set the fields that aren't already set in the reserve request
        (reserve_accept_message
         .set_field_value(fix.ClOrdID(), FIXApplication.get_next_clordid())
         .set_field_value(fix.HandlInst(), fix.HandlInst_MANUAL_ORDER_BEST_EXECUTION)
         .set_field_value(fix.OrdStatus(), fix.OrdStatus_NEW)
         .set_field_value(fix.Text(), f"Firm Up Order: {reserve_request_message.get_field_value(fix.OrderID())}")
         .set_field_value(fix.TimeInForce(), None)
         .set_field_value(fix.ExecBroker(), None)
         .set_field_value(fix.ClientID(), reserve_request_message.get_field_value(fix.ClOrdID()))
         )
        ServerApplication.create_order_message(MessageAction.NewOrder, reserve_accept_message, True, False)

    def send_reserve_reject_message(self, reserve_request_message: FIXMessage):
        reserve_reject_message = FIXMessage(reserve_request_message)
        # Only (un)set the fields that aren't already set in the reserve request
        (reserve_reject_message
         .set_field_value(fix.Currency(), FIXApplication.CURRENCY)
         .set_field_value(fix.ExecID(), FIXApplication.get_next_clordid())
         .set_field_value(fix.HandlInst(), fix.HandlInst_MANUAL_ORDER_BEST_EXECUTION)
         .set_field_value(fix.OrderQty(), "0")
         .set_field_value(fix.OrdStatus(), fix.OrdStatus_REJECTED)
         .set_field_value(fix.Text(), "Can Not Firm Up Order: for some odd reason...")
         .set_field_value(fix.ExecBroker(), None)
         .set_field_value(fix.ClientID(), None)
         .set_field_value(fix.ExecType(), fix.ExecType_REJECTED)
         )
        ServerApplication.create_order_message(MessageAction.RejectOrder, reserve_reject_message, True, False)

    @staticmethod
    def send_message(message: fix.Message):
        fix.Session.sendToTarget(message, ServerApplication.session_id)

    @staticmethod
    def side_str_to_fix(side: str) -> int:
        if side in OrderManager.SIDES:
            return OrderManager.SIDES[side]
        else:
            return 0

    @staticmethod
    def is_uuid_of_interest(uuid: str) -> bool:
        return str(uuid) in ServerApplication.uuids_of_interest

    @staticmethod
    def create_order_message(action: MessageAction, message: Series | FIXMessage,
                             send_message: bool = False, save_message: bool = False) -> fix.Message:
        if isinstance(message, Series):
            # Initiated from the UI
            fix_string = ServerApplication.create_fix_string_from_series(message)
            order_id = message['order_id']
            clordid = FIXApplication.get_next_clordid()
            FIXApplication.set_latest_clordid_per_order_id(order_id, clordid)
        else:
            # Initiated by receiving a message from the client
            order_id = message.get_field_value(fix.OrderID())
            fix_string = create_fix_string_from_dict(message.to_dict(), '|')

        # Weirdly enough BBG doesn't use tag41 and maintains the same tag11 thru 35=D/F's
        # if action == MessageAction.ChangeOrder or action == MessageAction.CancelOrder:
        #     latest_clordid = FIXApplication.get_latest_clordid_per_order_id(order_id)
        #     fix_string += f" 41={latest_clordid}"

        message = string_to_message(action.value, fix_string)

        if send_message:
            ServerApplication.send_message(message)

        if save_message:
            FIXApplication.set_latest_fix_message_per_order_id(order_id, FIXMessage.message_to_dict(message))

        return message

    @staticmethod
    def create_fix_string_from_series(order: Series) -> str:
        symbol = order['symbol']
        cusip = FIXApplication.KNOWN_SYMBOLS_BY_TICKER.get(symbol, f"??{symbol}??")

        side = ServerApplication.side_str_to_fix(order['side'])
        clordid = FIXApplication.get_next_clordid()
        order_id = order['order_id']
        order_price = float(order['price'])
        if order_price > 0.0:
            order_type = fix.OrdType_LIMIT
        else:
            order_type = fix.OrdType_MARKET
        fix_string = '|'.join([
            f"11={clordid}",
            f"15={FIXApplication.CURRENCY}",
            f"21={fix.HandlInst_MANUAL_ORDER_BEST_EXECUTION}",
            f"22={fix.IDSource_CUSIP}",
            f"37={order_id}",
            f"38={order['shares']}",
            f"44={order_price:.2f}",
            f"40={order_type}",
            f"48={cusip}",
            f"50={order['uuid']}",
            f"54={side}",
            f"55={symbol}",
            f"59={fix.TimeInForce_DAY}",
            f"60={get_utc_transactime()}",
            f"100={FIXApplication.EX_DESTINATION}",
            #            f"115={???}",  # OnBehalfOfCompID
            #            f"116={???}",  # OnBehalfOfSubID
            #            f"128={???}",  # DeliverToCompID
        ])

        return fix_string



from typing import List

import streamlit as st
from pandas import DataFrame

from order_manager import OrderManager
from fix_application import FIXApplication

GRID_KEY = 'grid'
VALID_TICKERS:List[str] = FIXApplication.KNOWN_SYMBOLS_BY_TICKER.keys()

def create_data_editor(orders_df: DataFrame) -> DataFrame:
    # the columns names must match the ones in the csv file
    edited_df = st.data_editor(orders_df, key=GRID_KEY,
                               height=420,
                               use_container_width=True,
                               num_rows='dynamic',
                               column_config={
                                   "order_id": st.column_config.TextColumn(
                                   ),
                                   "uuid":st.column_config.NumberColumn(
                                       format="%d",
                                   ),
                                   "side": st.column_config.SelectboxColumn(
                                       options=OrderManager.SIDES
                                   ),
                                   "price": st.column_config.NumberColumn(
                                       format="$ %.2f",
                                   ),
                                   "symbol":st.column_config.SelectboxColumn(
                                       options=VALID_TICKERS,
                                   ),
                               },
                               hide_index=True,
                               )
    return edited_df


def process_edited_added_rows(orders_df: DataFrame, edited_df: DataFrame) -> None:
    edited_rows = st.session_state[GRID_KEY]["edited_rows"]
    added_rows = st.session_state[GRID_KEY]["added_rows"]
    initial_change_count = valid_change_count = len(edited_rows) + len(added_rows)
    edited_rows = st.session_state[GRID_KEY]["edited_rows"]
    for index, edited_columns in edited_rows.items():
        if 'order_id' in edited_columns or 'uuid' in edited_columns:
            current_order_id = orders_df.iloc[index]['order_id']
            st.error(
                f"Can't change order_id and UUID. Ignoring the changed row with current order_id:{current_order_id}")
            edited_df.iloc[int(index)] = orders_df.iloc[int(index)]
            valid_change_count -= 1

    if valid_change_count:
        order_manager.orders_df = edited_df
        order_manager.save_orders()

        order_manager.save_order_change_instructions(st.session_state[GRID_KEY])
        st.success('Orders saved/sent successfully!')
    else:
        if initial_change_count != 1:
            st.warning('No valid order left to save/send')


def main(order_manager: OrderManager) -> None:
    st.set_page_config(layout="wide")

    orders_df = order_manager.orders_df
    st.title('OMS Orders Management')
    edited_df = create_data_editor(orders_df)

    if st.button('Save/Send Changes'):
        # st.write(st.session_state[GRID_KEY])

        if len(st.session_state[GRID_KEY]["deleted_rows"]):
            st.error("Orders can't be deleted, just mark them as inactive. Ignoring all changes")
        else:
            valid_row_count = (len(st.session_state[GRID_KEY]["edited_rows"]) +
                               len(st.session_state[GRID_KEY]["added_rows"]))
            if valid_row_count == 0:
                st.warning("No change to save/send")
            else:
                process_edited_added_rows(orders_df, edited_df)

        st.button('Close')

    # TODO: add a button to stop server/client app, reset the state, and restart


if __name__ == "__main__":
    order_manager = OrderManager()
    main(order_manager)



import threading
from datetime import datetime, timedelta
from typing import Dict, Union, Set, Any, Optional, List

import quickfix as fix
from quickfix import Message

self_lock = threading.Lock()


class FIXMessage():
    def __init__(self, message: Dict[str, str] | Message | Optional['FIXMessage'] = None):
        if message:
            if isinstance(message, Message):
                self.message_dict = FIXMessage.message_to_dict(message)
            elif isinstance(message, FIXMessage):
                self.message_dict = dict(message.message_dict)
            else:
                self.message_dict = dict(message)
        else:
            self.message_dict = dict()

    def get_field_value(self, fix_field_obj: Any) -> str:
        fix_key_as_str = str(fix_field_obj.getField())
        field_value = self.message_dict.get(fix_key_as_str, None)

        return field_value

    def set_field_value(self, fix_field_obj: Any, field_value: str | None) -> Optional['FIXMessage']:
        fix_key_as_str = str(fix_field_obj.getField())
        if field_value is None:
            if fix_key_as_str in self.message_dict:
                del self.message_dict[fix_key_as_str]
        else:
            self.message_dict[fix_key_as_str] = field_value

        return self

    def to_dict(self):
        return dict(self.message_dict)

    def __str__(self):
        return '|'.join([f"{k}={v}" for k, v in self.message_dict.items()])

    @staticmethod
    def message_to_dict(message: Message) -> Dict[str, str]:
        message_dict: Dict[str, str] = {}
        string_with_ctrla = message.toString()
        kv_pairs = string_with_ctrla.split('\x01')
        for kv_pair in kv_pairs:
            if '=' in kv_pair:
                key, value = kv_pair.split('=')
                message_dict[key] = value

        return message_dict

    @staticmethod
    def message_to_string(message: fix.Message | Dict[str, str]) -> str:
        if isinstance(message, fix.Message):
            string_with_ctrla = message.toString()
            string = string_with_ctrla.replace('\x01', '|')
        elif isinstance(message, dict):
            string = '|'.join([f"{k}={v}" for k, v in message.items()])

        return string


class FIXApplication(fix.Application):
    SESSION_LEVEL_TAGS: Set[str] = {str(tag) for tag in [8, 9, 10, 34, 35, 49, 52, 56]}
    KNOWN_SYMBOLS_BY_TICKER: Dict[str, str] = {
        "BOOM": "23291C10",
        "CAKE": "16307210",
        "FUN": "15018510",
        "HEINY": "42301230",
        "HOG": "41282210",
        "LUV": "84474110",
        "PLAY": "23833710",
        "ROCK": "37468910",
        "ZEUS": "68162K10",
        "ZVZZT": "0ZVZZT88",
    }
    KNOWN_SYMBOLS_BY_CUSIP: Dict[str, str] = {v: k for k, v in KNOWN_SYMBOLS_BY_TICKER.items()}

    EXEC_BROKER = 'ITGI'
    LAST_MARKET = 'ITGI'
    EX_DESTINATION = 'US'
    CURRENCY = 'USD'

    latest_clordid_per_order_id: Dict[str, str] = {}
    latest_fix_message_per_order_id: Dict[str, FIXMessage] = {}
    base_clordid = datetime.now().strftime("%Y%m%d%H%M%S")
    current_clordid = 0

    def onCreate(self, session_id):
        pass

    def onLogon(self, session_id):
        print(f"Session {session_id} logged on.")

    def onLogout(self, session_id):
        print(f"Session {session_id} logged out.")

    def toAdmin(self, message, session_id):
        pass

    def fromAdmin(self, message, session_id):
        pass

    def toApp(self, message, session_id):
        pass

    def fromApp(self, message, session_id):
        print(f"Received message:{FIXMessage(message)}")

    @staticmethod
    def get_next_clordid():
        FIXApplication.current_clordid += 1
        return f"{FIXApplication.base_clordid}{FIXApplication.current_clordid:06}"

    @staticmethod
    def set_latest_clordid_per_order_id(order_id: str, clordid: Union[str, None]) -> None:
        if clordid:
            FIXApplication.latest_clordid_per_order_id[order_id] = clordid
        else:
            # since clordid is None, remove entry if exists
            if order_id in FIXApplication.latest_clordid_per_order_id:
                FIXApplication.latest_clordid_per_order_id.pop(order_id)

    @staticmethod
    def get_latest_clordid_per_order_id(order_id: str) -> Union[str, None]:
        return FIXApplication.latest_clordid_per_order_id.get(order_id, None)

    @staticmethod
    def set_latest_fix_message_per_order_id(order_id: str, message: Union[Dict[str, str], FIXMessage, None]) -> None:
        with self_lock:
            if message:
                if isinstance(message, dict):
                    message = FIXMessage(message)
                FIXApplication.latest_fix_message_per_order_id[order_id] = message
            else:
                # since message is None, remove entry if exists
                if order_id in FIXApplication.latest_fix_message_per_order_id:
                    FIXApplication.latest_fix_message_per_order_id.pop(order_id)

    @staticmethod
    def get_latest_fix_message_per_order_id(order_id: str, issue_error: bool = True) -> FIXMessage | None:
        with self_lock:
            latest_message = FIXApplication.latest_fix_message_per_order_id.get(order_id, None)
            if latest_message:
                return latest_message
            else:
                if issue_error:
                    print(f"ERROR: Can't find a FIX message for order_id:{order_id}")
                return None


def string_to_message(message_type: int, fix_string: str, separator: str = '|') -> Message:
    message = fix.Message()

    header = message.getHeader()
    header.setField(fix.BeginString(fix.BeginString_FIX42))
    header.setField(fix.MsgType(message_type))

    tag_value_pairs = fix_string.split(separator)
    for pair in tag_value_pairs:
        try:
            tag, value = pair.split("=")
        except Exception as e:
            print(f"ERROR! Can't extract key/value from:{pair} with exception:{e}")
            continue
        if tag not in FIXApplication.SESSION_LEVEL_TAGS:
            message.setField(int(tag), value)

    return message


def create_fix_string_from_dict(message: Dict[str, str], separator: str = '|') -> str:
    tag_value_pairs: List[str] = []
    for tag, value in message.items():
        if tag not in FIXApplication.SESSION_LEVEL_TAGS:
            if tag == fix.TransactTime().getField():  # tag60
                value = get_utc_transactime()
            tag_value_pairs.append(f"{tag}={value}")

    fix_string = separator.join(tag_value_pairs)

    return fix_string


def get_header_field_value(msg, fobj) -> Union[str, None]:
    if msg.getHeader().isSetField(fobj.getField()):
        msg.getHeader().getField(fobj)
        return fobj.getValue()
    else:
        return None


def timestamp() -> str:
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]


def get_utc_transactime(offset_in_secs: int = 0) -> str:
    utc_time = datetime.utcnow()
    if offset_in_secs:
        utc_time += timedelta(seconds=offset_in_secs)

    return utc_time.strftime("%Y%m%d-%H:%M:%S.%f")


def log(msg_type: str, message: str | Dict[str, str] | fix.Message | FIXMessage | None = None,
        pre_timestamp: str = '') -> None:
    if message == None:
        message = ''
    elif isinstance(message, dict) or isinstance(message, fix.Message):
        message = FIXMessage(message)

    print(f"{pre_timestamp}{timestamp()} {msg_type:<11}: {message}")
